import os
import requests
from tqdm.auto import tqdm
from tqdm.utils import CallbackIOWrapper

from .base import Base
from .utils import check, dataset
class DatasetManager(Base):

    def upload_file(self, directory_id=None, file_path=None, file_name=None, description="", lineage=None, replace=None):
        """_summary_

        Args:
            directory_id (_type_, optional): _description_. Defaults to None.
            file_path (_type_, optional): _description_. Defaults to None.
            description (str, optional): _description_. Defaults to "".
            lineage (_type_, optional): _description_. Defaults to None.

        Raises:
            Exception: _description_
            Exception: _description_
            Exception: _description_
            Exception: _description_
            Exception: _description_
            Exception: _description_

        Returns:
            _type_: _description_
        """
        if type(file_path) != str :
            raise Exception(f"file_path expect `str` but got {type(file_path)}, or please input `file_path='<path>'`")
        elif not os.path.exists(file_path):
            raise Exception(f"file {os.path.abspath(file_path)} not found")
        if directory_id == None:
             raise Exception("please input directory, like `directory=<id>`")
        _res = requests.get(f"{self._discovery_api}/directory/{directory_id}",  headers=self._jwt_header)
        if _res.status_code == 404:
            raise Exception(f"directory {directory_id} not found or your dose not has permission to access directory")
        
        if file_name:
            f_name = file_name
        else:
            f_name = os.path.basename(file_path)
            
        replace = self._check_fileExists(directory_id, f_name) if replace is None else replace
        
        if description == "":
            description = f"file {f_name}"
            
        _res = requests.post(f"{self._discovery_api}/file/", headers=self._jwt_header,
                             data={
                                 'name': f_name,
                                 'description': description,
                                 'directory': int(directory_id),
                                 'size': os.path.getsize(file_path),
                                 'replace': replace
                             }
                )
        if not _res.status_code in [200, 201]:
            txt = ""
            if _res.status_code < 500:
                txt = _res.json()
            raise Exception(f"{_res.status_code}, {txt}")
        data = _res.json()
        if self._internal:
            data['url'] = self._replace_minio_api(data['url'])
        file_size = os.stat(file_path).st_size
        with open(file_path, 'rb') as f:
            with tqdm(total=file_size, unit="B", unit_scale=True, unit_divisor=1024) as t:
                wrapped_file = CallbackIOWrapper(t.update, f, "read")
                _res = requests.put(data['url'], data=wrapped_file)
        if _res.status_code == 200:
            _uploaded = True
        else:
            if self._verbose: print(_res.content.decode())
            raise Exception("Some thing wrong with datastore")
        data.update({
            'uploaded': _uploaded,
        })
        data = {key: value for key, value in data.items() if key in ['uploaded', 'id', 'msg', 'path']}
        
        if lineage != None:
            check.check_lists_int(data=lineage, dataName="lineage")
            r = requests.post(f"{self._discovery_api}/file/{data['id']}/setLineage/", 
                                headers=self._jwt_header,
                                json={
                                    'lineage': lineage
                                }
            )
            try:
                check.check_http_status_code(r)
            except Exception as e:
                data.update({
                    'lineage_msg': str(e),
                    'lineage': False
                })
            else:
                data.update({
                    'lineage': "created",
                    'lineage': True
                })
            
        
        return data
    
    def download_file(self, file_id=None, download_path=None):
        """_summary_

        Args:
            file_id (_type_, optional): _description_. Defaults to None.
            download_path (_type_, optional): _description_. Defaults to None.

        Raises:
            Exception: _description_
            Exception: _description_
            Exception: _description_
            Exception: _description_

        Returns:
            _type_: _description_
        """
        if file_id == None:
             raise Exception("please input file_id")
        _res = requests.get(f"{self._discovery_api}/file/{file_id}/",  headers=self._jwt_header)
        if _res.status_code != 200:
            raise Exception(f"File response code {_res}")
        meta = _res.json()
        _res = requests.get(f"{self._discovery_api}/file/{file_id}/download/",  headers=self._jwt_header)
        if _res.status_code != 200:
            raise Exception("some thing wrong in datastore")
            
        _url = _res.json()['url']
        if self._internal:
            _url = self._replace_minio_api(_url)
            
        _res = requests.get(_url)
        f_path = os.path.join(self._tmp_path, meta['name'])
        if download_path != None:
            if (not os.path.exists(download_path)) or (not os.path.isdir(download_path)):
                raise Exception("`download_path` is not existx or is not directory")
            else:
                f_path = os.path.join(self._tmp_path, meta['name'])
            
        
        total_size_in_bytes= int(_res.headers.get('content-length', 0))
        if self._verbose: print(f"total size : {total_size_in_bytes} B")
        block_size = 1024 #1 Kibibyte
        progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
        
        with open(f_path, 'wb') as f:
            for data in _res.iter_content(block_size):
                progress_bar.update(len(data))
                f.write(data)
        progress_bar.close()
        
        meta.update({
            'download_sucess': os.path.exists(f_path),
            'f_path': f_path
        })
        return meta
    
    def createDataset(self, name=None, tags=None):
        """_summary_

        Args:
            name (_type_, optional): _description_. Defaults to None.
            tags (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        check.check_type(variable=name, variableName="name", dtype=str)
        check.check_type(variable=tags, variableName="tags", dtype=list, child=str)
        
        tags_ids = dataset.get_tag_ids(base_uri=self._base_discovery_api, tags=tags, header=self._jwt_header)
        
        r = requests.post(f"{self._base_discovery_api}/api/v2/dataset/", headers=self._jwt_header,
                          json={
                              'name': name,
                              'title': name,
                              'tag': tags_ids
                          }
        )
        check.check_http_status_code(response=r)
        return r.json()
    
    def createDirectory(self, directory_id=None, name=None, description=None):
        """_summary_

        Args:
            directory_id (_type_, optional): _description_. Defaults to None.
            name (_type_, optional): _description_. Defaults to None.
            description (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        check.check_type(variable=directory_id, variableName="directory_id", dtype=int)
        check.check_type(variable=name, variableName="name", dtype=str)
        check.check_type(variable=description, variableName="description", dtype=str)
        
        r = requests.post(f"{self._base_discovery_api}/api/v2/directory/", headers=self._jwt_header,
                          json={
                              "name": name,
                              "description": description,
                              "parent_dir": directory_id
                          }
        )
        check.check_http_status_code(response=r)
        return r.json()


def check_http_status_code(response):
    if response.status_code >= 300:
        txt = response.json() if response.status_code < 500 else " "
        raise Exception(f"Some thing wrong {response.status_code}, {txt}")


import requests
from . import check
import dask.dataframe as dd
from operator import itemgetter
import hashlib

def md5hash(x):
    return hashlib.md5(x.encode('utf-8')).hexdigest()

def get_list_datanode(meta, index, headers, base_uri, storage_options, bucketName):
    version_file_ids = meta.get('context', {}).get('version_file_id', [])
    version_file_ids = version_file_ids.values()
    version_file_ids = sorted(version_file_ids, key=itemgetter('timestamp'), reverse=False)
    
    if index < len(version_file_ids):
        version_file_id = version_file_ids[index]
        print(f"version_file_id : {version_file_id}")
        _res = requests.get(f"{base_uri}/api/v2/file/{version_file_id.get('file_id')}/", headers=headers)
        check.check_http_status_code(response=_res)
        meta = _res.json()
        return dd.read_parquet(f"s3://{bucketName}/{meta['s3_key']}", storage_options=storage_options)
    else:
        # index out of range
        raise Exception(f"list_datanode is out of range for index {index}")

def get_change_data(
        ddf_current: dd.DataFrame, 
        ddf_previous: dd.DataFrame,
    ) -> dd.DataFrame:

    ddf_current['_hash'] = ddf_current.astype(str).values.sum(axis=1)
    ddf_current['_hash'] = ddf_current['_hash'].apply(md5hash, meta=('_hash', 'object'))   

    if ddf_previous is not None:
        # already have landing of this table
        ddf_previous['_hash'] = ddf_previous.astype(str).values.sum(axis=1)
        ddf_previous['_hash'] = ddf_previous['_hash'].apply(md5hash, meta=('_hash', 'object'))
        previous_hash_index = ddf_previous['_hash'].unique().compute()        
        ddf_change = ddf_current[~ddf_current['_hash'].isin(previous_hash_index)]
    else:
        ddf_change = ddf_current
        
    ddf_change = ddf_change.drop(columns=['_hash'])

    return ddf_change


import os
import requests
import time

from minio import Minio

defind_clear_output = False
try:
    from IPython.display import clear_output
    defind_clear_output = True
except:
    pass

discovery = os.environ.get('discovery_url', "https://discovery.data.storemesh.com")

_base_discovery_api = os.environ.get('BASE_DISCOVERY_API', "https://api.discovery.data.storemesh.com")
_internal_base_discovery_api = os.environ.get('INTERNAL_BASE_DISCOVERY_API',"http://discovery-backend:8000")

_external_api_minio = os.environ.get('external_api_minio',"")
_internal_api_minio = os.environ.get('internal_api_minio',"")

_base_object_storage_url = os.environ.get('base_minio_url', "api.minio.data.storemesh.com")
bucket_name = os.environ.get('bucketname', 'dataplatform')

class Base:
    def __init__(self, token=None, verbose=True, dataplatform_api_uri=_base_discovery_api, object_storage_uri=_base_object_storage_url):
        """_summary_

        Args:
            token (str, required): token get from discovery. Defaults to None.
            verbose (bool, optional): display logging. Defaults to True.
            dataplatform_api_uri (str, optional): dataplatform uri. Defaults to "https://api.discovery.data.storemesh.com".

        Raises:
            Exception `Please enter your token from dsmOauth`: Invalid token
            Exception `Can not connect to DataPlatform`: Some thing wrong connection with dataplatform
            Exception `Can not get objectstorage user`: Some thing wrong connection with objectstorage
        """
        
        self._verbose = verbose
        _internal = os.environ.get('internal', None) == "true"
        self._internal = _internal
        _base_minio_url = _base_object_storage_url
        if self._internal:
            base_discovery_api = _internal_base_discovery_api
        else:
            base_discovery_api = dataplatform_api_uri
            _base_minio_url = object_storage_uri
        self._base_discovery_api = base_discovery_api
        self._discovery_api = f"{base_discovery_api}/api/v2"
        
        if token is None:
            print(f"Please get token from {self._base_discovery_api.replace('api.','')}")
            token = input("Your Token : ")
            if defind_clear_output:
                time.sleep(2)
                clear_output()
        if token in [None, '']:
            raise Exception('Please enter your token from discovery')
        self._jwt_header = {
            'Authorization': f'Bearer {token}'
        }
        _res = requests.get(f"{base_discovery_api}/api/v2/account/me/", headers=self._jwt_header)
        if _res.status_code != 200:
            txt = _res.json() if _res.status_code < 500 else " "
            raise Exception(f"Can not connect to DataPlatform, {txt}")
        self.token = token
        self._tmp_path = 'dsm.tmp'
        os.makedirs(self._tmp_path, exist_ok=True)
        _res = requests.get(f"{base_discovery_api}/api/minio/minio-user/me/", headers=self._jwt_header)
        if _res.status_code != 200:
            raise Exception("Can not get objectstorage user")
        data = _res.json()
        self._minio_access = data['access']
        self._minio_secret = data['secret']

        self.client = Minio(
            _base_minio_url,
            access_key=self._minio_access,
            secret_key=self._minio_secret,
            secure=not _internal
        )
        _scheme = "http" if self._internal else "https"
        self._storage_options = {
            'key': self._minio_access,
            'secret': self._minio_secret,
            'client_kwargs':{
                'endpoint_url': f"{_scheme}://{_base_minio_url}"
            }
        }
        if self._verbose: print("Init DataNode sucessful!")
        
    def _replace_minio_api(self, url):
        return url.replace('https', 'http').replace(_external_api_minio, _internal_api_minio)
    
    def _check_fileExists(self, directory, name):
        _res = requests.get(f"{self._discovery_api}/directory/{directory}/fileExists/?filename={name}", headers=self._jwt_header)
        if _res.status_code == 200:
            replace = False
        elif _res.status_code == 302:
            replace = input(f"File {name} alrady exists, do you want to replace y/n : ").replace(' ', '').lower() == "y"
        else:
            status = _res.json() if _res.status_code < 500 else ""
            raise Exception(f"check file exists error {status}")
        return replace
    
    def _check_fileExists_no_ask(self, directory, name):
        _res = requests.get(f"{self._discovery_api}/directory/{directory}/fileExists/?filename={name}", headers=self._jwt_header)
        if _res.status_code == 200:
            is_exist = False
        elif _res.status_code == 302:
            is_exist = True
        else:
            status = _res.json() if _res.status_code < 500 else ""
            raise Exception(f"check file exists error {status}")
        return is_exist